@@ -77,10 +77,11 @@ LOG = logging.getLogger(__name__)
 
 # reference to the ASN.1 Data view and to the process variables
 TYPES = None
-VARIABLES = None
+VARIABLES = {}
 # List of output signals and procedures
-OUT_SIGNALS = None
-PROCEDURES = None
+OUT_SIGNALS = []
+PROCEDURES = []
+INNER_PROCEDURES = []
 
 # lookup table for mapping SDL operands with the corresponding Ada ones
 OPERANDS = {'plus': '+', 'mul': '*', 'minus': '-', 'or': 'or',
@@ -111,7 +112,7 @@ def get_type_of_parent(identifier):
             for index in range(1, len(identifier)):
                 # TODO find type...
                 LOG.warning('** INCOMPLETE FEATURE ** ')
-            kind = current_type['Kind']
+            kind = current_type['Kind'].replace('-', '_')
     return kind, name
 
 
@@ -144,7 +145,7 @@ def write_statement(param, newline):
             cast = 'Boolean'
         string = "{cast}'Image({s})".format(cast=cast, s=string)
     else:
-        error = ('Unsupported parameter in write call' +
+        error = ('Unsupported parameter in write call ' +
                 param.var.inputString)
         LOG.error(error)
         raise TypeError(error)
@@ -180,55 +181,67 @@ def output_statement(output):
                 code.extend(stmts)
                 local_decl.extend(local)
             continue
-        if output.kind == 'output':
-            out_sig = OUT_SIGNALS.get(signal_name.lower())
+        proc, out_sig = None, None
+        try:
+            out_sig, = [sig for sig in OUT_SIGNALS
+                        if sig['name'] == signal_name.lower()]
+        except ValueError:
+            # Not an output, try if it is an external procedure
+            try:
+                out_sig, = [sig for sig in PROCEDURES
+                            if sig.inputString.lower() == signal_name.lower()]
+            except ValueError:
+                # Not external? Must be an inner procedure then.
+                # otherwise the parser would have barked
+                proc, = [sig for sig in INNER_PROCEDURES
+                         if sig.inputString.lower() == signal_name.lower()]
+        if out_sig:
+            list_of_params = []
+            for idx, param in enumerate(out.get('params') or []):
+                param_direction = 'in'
+                try:
+                    # If it is an output, there is a single parameter
+                    param_type = out_sig['type']
+                except TypeError:
+                    # Else if it is a procedure, get the type
+                    param_type = out_sig.fpar[idx]['type']
+                    param_direction = out_sig.fpar[idx]['direction']
+
+                p_code, p_id, p_local = decipher_expression(param)
+                code.extend(p_code)
+                local_decl.extend(p_local)
+                # Create a temporary variable for input parameters only
+                if param_direction == 'in':
+                    tmp_id = out['tmpVars'][idx]
+                    local_decl.append('tmp{idx} : aliased asn1Scc{oType};'.format(
+                        idx=tmp_id, oType=param_type))
+                    code.append('tmp{idx} := {p_id};'.format(
+                        idx=tmp_id, p_id=p_id))
+                    list_of_params.append("tmp{idx}'access".
+                            format(idx=out['tmpVars'][idx]))
+                else:
+                    # Output parameters - no need for a temp variable
+                    list_of_params.append("{var}'access".format(var=p_id))
+            if list_of_params:
+                code.append('{RI}({params});'.format(
+                    RI=out['outputName'], params=', '.join(list_of_params)))
+            else:
+                code.append('{RI};'.format(RI=out['outputName']))
         else:
-            out_sig = PROCEDURES.get(signal_name.lower())
-        if not out_sig:
-            error = 'Signal or procedure not defined:' + str(signal_name)
-            LOG.error(error)
-            raise NameError(error)
-        nb_in_params = len(out_sig['paramsInOrdered'])
-        list_of_params = []
-        for idx, param in enumerate(out.get('params') or []):
-            # Determine if it is an IN or OUT type
-            select = ('paramsOutOrdered', 'out', idx - nb_in_params) if (
-                    idx >= nb_in_params) else ('paramsInOrdered', 'in', idx)
-
-            # Get parameter name and type from the Interface view:
-            param_name = out_sig[select[0]][select[2]]
-            param_type = out_sig[select[1]][param_name]['type']
-
-            # Check the type of the parameter - set it if missing
-            # (TODO: do it in the parser)
-            if param.exprType.get('Kind') == 'UnknownType':
-                param.exprType = {'Kind': 'ReferenceType',
-                        'ReferencedTypeName': param_type}
-
-            p_code, p_id, p_local = decipher_expression(param)
-            code.extend(p_code)
-            local_decl.extend(p_local)
-            # Create a temporary variable for this parameter
-            tmp_id = out['tmpVars'][idx]
-            local_decl.append('tmp{idx} : aliased asn1Scc{oType};'.format(
-                idx=tmp_id, oType=param_type))
-            code.append('tmp{idx} := {p_id};'.format(
-                idx=tmp_id, p_id=p_id))
-            list_of_params.append("tmp{idx}'access".
-                    format(idx=out['tmpVars'][idx]))
-        if list_of_params:
-            code.append('{RI}({params});'.format(
-                RI=out['outputName'], params=', '.join(list_of_params)))
-            # Assign back the out parameters to the local variables
-            if len(out_sig['paramsOutOrdered']) > 0:
-                for (idx, parameter) in ((idx, p)
-                        for idx, p in enumerate(out['params'])
-                        if idx >= nb_in_params):
-                    _, p_id, __ = decipher_expression(parameter)
-                    code.append('{p_id} := tmp{idx};'.format(
-                        p_id=p_id, idx=out['tmpVars'][idx]))
-        else:
-            code.append('{RI};'.format(RI=out['outputName']))
+            # inner procedure call
+            list_of_params = []
+            for param in (out.get('params') or []):
+                p_code, p_id, p_local = decipher_expression(param)
+                code.extend(p_code)
+                local_decl.extend(p_local)
+                # no need to use temporary variables, we are in pure Ada
+                list_of_params.append(p_id)
+            if list_of_params:
+                code.append('{proc}({params});'.format(
+                    proc=proc.inputString,
+                    params=', '.join(list_of_params)))
+            else:
+                code.append('{proc};'.format(proc.inputString))
     return code, local_decl
 
 
@@ -343,7 +356,7 @@ def decipher_primary_id(primaryId):
                     # needed to build the Ada expression
                     # (Won't work for embedded types - FIXME)
                     exp_typename = (exp.exprType.get('ReferencedTypeName') or
-                            exp.exprType['Kind'])
+                            exp.exprType['Kind']).replace('-', '_') 
                     if exp_type['Kind'] != 'ChoiceType':
                         error = '{} is not a CHOICE'.format(exp.inputString)
                         LOG.error(error)
@@ -435,7 +448,7 @@ def decipher_primary(primary):
                 primary.choiceItem, primary.primType)
     elif primary.kind == 'emptyString':
         ada_string = 'asn1Scc{typeRef}_Init'.format(
-                typeRef=primary.primType['ReferencedTypeName'])
+              typeRef=primary.primType['ReferencedTypeName'].replace('-', '_'))
     elif primary.kind == 'stringLiteral':
         pass
     elif primary.kind == 'ifThenElse':
@@ -456,7 +469,8 @@ def decipher_if_then_else(ifThenElse, resType):
     # FIXME: resType may be wrong if declaration embedded in SEQUENCE
     # TODO: find a non-conflicing naming convention for tmp variable
     local_decl = ['tmp{idx} : asn1Scc{resType};'.format(
-        idx=ifThenElse['tmpVar'], resType=resType.get('ReferencedTypeName') or
+        idx=ifThenElse['tmpVar'],
+        resType=resType.get('ReferencedTypeName').replace('-', '_') or
         'ERROR!')]
     if_stmts, if_str, if_local = decipher_expression(ifThenElse['if'])
     then_stmts, then_str, then_local = decipher_expression(ifThenElse['then'])
@@ -483,7 +497,7 @@ def decipher_sequence(seq, seqType):
     ''' Return Ada string for an ASN.1 SEQUENCE '''
     LOG.debug('Sequence: ' + str(seq) + str(seqType))
     ada_string = 'asn1Scc{seqType}\'('.format(
-            seqType=seqType['ReferencedTypeName'])
+            seqType=seqType['ReferencedTypeName'].replace('-', '_'))
     stmts = []
     local_decl = []
     sep = ''
@@ -513,16 +527,18 @@ def decipher_sequence_of(seqof, seqofType):
     asn_type = TYPES[typename]['type']
     min_size = asn_type['Min']
     max_size = asn_type['Max']
-    ada_string = 'asn1Scc{seqofType}\'('.format(seqofType=typename)
+    ada_string = 'asn1Scc{seqofType}\'('.format(
+                                seqofType=typename.replace('-', '_'))
     if min_size == max_size:
         # Fixed-length array - no need to set the Length field
         ada_string += 'Data => asn1Scc{seqofType}_array\'('.format(
-                seqofType=typename)
+                seqofType=typename.replace('-', '_'))
     else:
         # Variable-length array
         ada_string += (
                 'Length => {length}, Data => asn1Scc{seqofType}_array\'('
-                .format(seqofType=typename, length=len(seqof)))
+                .format(seqofType=typename.replace('-', '_'),
+                        length=len(seqof)))
     for i in xrange(len(seqof)):
         if seqof[i].primType is None:
             # Embedded type (e.g. SEQUENCE OF SEQUENCE {...}):
@@ -548,9 +564,10 @@ def decipher_choice(choice, choiceType):
                     choiceValue=choice['choice'])}
 
     stmts, choice_str, local_decl = decipher_expression(choice['value'])
-
+    actual_type = choiceType.get('ReferencedTypeName') or choiceType['Kind']
+    actual_type = actual_type.replace('-', '_')
     ada_string = 'asn1Scc{cType}_{opt}_set({expr})'.format(
-            cType=choiceType.get('ReferencedTypeName') or choiceType['Kind'],
+            cType=actual_type,
             opt=choice['choice'],
             expr=choice_str)
     return stmts, ada_string, local_decl
@@ -654,11 +671,12 @@ def transition(tr):
         if tr.terminator.label:
             code.append('<<{label}>>'.format(
                 label=tr.terminator.label.inputString))
-        if(tr.terminator.kind == 'next_state' and
-                tr.terminator.inputString.strip() != '-'):
-            # discard the "-" state (remain in the same state)
-            code.append('state := {nextState};'.format(
-                nextState=tr.terminator.inputString))
+        if tr.terminator.kind == 'next_state':
+            if tr.terminator.inputString.strip() != '-':
+                # discard the dash state (remain in the same state)
+                code.append('state := {nextState};'.format(
+                             nextState=tr.terminator.inputString))
+            # In any case, return to avoid code of floating labels
             code.append('return;')
         elif tr.terminator.kind == 'join':
             code.append('goto {label};'.format(
@@ -669,6 +687,73 @@ def transition(tr):
     return code, local_decl
 
 
+def floating_label(label):
+    ''' Generate the code for a floating label (Ada label + transition) '''
+    code = []
+    local_decl = []
+    # Add the traceability information
+    code.extend(traceability(label))
+    code.append('<<{label}>>'.format(label=label.inputString))
+    if label.transition:
+        code_trans, local_trans = transition(label.transition)
+    code.extend(code_trans)
+    local_decl.extend(local_trans)
+    return code, local_decl
+
+
+def inner_procedure(proc):
+    ''' Generate the code for a procedure - does not support states '''
+    code = []
+    local_decl = []
+    # TODO: Update the global list of procedures
+    # with procedure defined inside the current procedure
+    # Not critical: the editor forbids procedures inside procedures
+
+    # Build the procedure signature
+    pi_header = 'procedure {proc_name}'.format(proc_name=proc.inputString)
+    if proc.fpar:
+        pi_header += '('
+        params = []
+        for fpar in proc.fpar:
+            params.append('l_{name}: in{out} asn1Scc{ptype}'.format(
+                    name = fpar.get('name'),
+                    out = ' out' if fpar.get('direction') == 'out' else '',
+                    ptype = fpar.get('type')))
+        pi_header += ';'.join(params)
+        pi_header += ')'
+
+    local_decl.append(pi_header + ';')
+
+
+    # Generate the code for the procedure itself (unless it's external):
+    # local variables and code of the START transition
+    if not proc.external:
+        # Recursively generate the code for inner-defined procedures
+        for inner_proc in proc.inner_procedures:
+            inner_code, inner_local = inner_procedure(inner_proc)
+            local_decl.extend(inner_local)
+            code.extend(inner_code)
+        code.append(pi_header + ' is')
+        for var_name, var_type in proc.variables.viewitems():
+            code.append('l_{name} : asn1Scc{sort};'.format(
+                name=var_name, sort=var_type))
+        tr_code, tr_decl = transition(proc.start.transition)
+        # Generate code for the floating labels
+        code_labels = []
+        for label in proc.floating_labels:
+            code_label, label_decl = floating_label(label)
+            code_labels.extend(code_label)
+            tr_decl.extend(label_decl)
+        code.extend(tr_decl)
+        code.append('begin')
+        code.extend(tr_code)
+        code.extend(code_labels)
+        code.append('end {procName};'.format(procName=proc.inputString))
+    code.append('\n')
+
+    return code, local_decl
+
+
 def format_ada_code(stmts):
     ''' Indent properly the Ada code '''
     indent = 0
@@ -692,24 +777,28 @@ def format_ada_code(stmts):
 def generate(process):
     ''' Generate Ada code '''
     process_name = process.processName
-    global VARIABLES
-    VARIABLES = process.variables
+    VARIABLES.update(process.variables)
     global TYPES
-    TYPES = process.dataView
-    global OUT_SIGNALS
-    global PROCEDURES
+    TYPES = process.dataview
     # Ada is case insensitive => lower cases
-    OUT_SIGNALS = {sig.lower(): defn for sig, defn in
-            process.outputSignals.viewitems()}
-    PROCEDURES = {p.lower(): defn for p, defn in
-            process.procedures.viewitems()}
-    filename = process_name + ".adb"
-    LOG.info('Generating Ada code for process' + str(process_name))
+    #OUT_SIGNALS.update({sig.lower(): defn for sig, defn in
+    #        process.outputSignals.viewitems()})
+    del OUT_SIGNALS[:]
+    del PROCEDURES[:]
+    del INNER_PROCEDURES[:]
+    OUT_SIGNALS.extend(process.output_signals)
+    PROCEDURES.extend(process.procedures)
+
+    #PROCEDURES.update({p.lower(): defn for p, defn in
+    #        process.procedures.viewitems()})
+    INNER_PROCEDURES.extend(process.inner_procedures)
+
+    LOG.info('Generating Ada code for process ' + str(process_name))
 
     # Generate the code to declare process-level variables
     process_level_decl = []
     for var_name, var_type in VARIABLES.iteritems():
-        process_level_decl.append('l_{n} : asn1Scc{t};'.format(
+        process_level_decl.append('l_{n} : aliased asn1Scc{t};'.format(
             n=var_name, t=var_type))
 
     # Add the process states list to the process-level variables
@@ -721,13 +810,15 @@ def generate(process):
     # Add the declaration of the runTransition procedure
     process_level_decl.append('procedure runTransition(trId: Integer);')
 
+
     # Generate the code of the start transition:
     start_transition = ['begin']
     start_transition.append('runTransition(0);')
 
     mapping = {}
     # Generate the code for the transitions in a mapping input-state
-    for input_signal in process.inputSignals.viewkeys():
+    input_signals = [sig['name'] for sig in process.input_signals]
+    for input_signal in input_signals: #process.inputSignals.viewkeys():
         mapping[input_signal] = {}
         for state_name, input_symbols in process.mapping.viewitems():
             if state_name != 'START':
@@ -740,7 +831,10 @@ def generate(process):
     asn1_modules = '\n'.join(['with {dv};\nuse {dv};'.format(
         dv=dv.replace('-', '_'))
         for dv in process.asn1Modules])
-    taste_template = ['''with System.IO;
+    taste_template = ['''\
+-- This file was generated automatically: DO NOT MODIFY IT !
+
+with System.IO;
 use System.IO;
 
 {dataview}
@@ -754,19 +848,43 @@ use Interfaces;
 package body {process_name} is'''.format(process_name=process_name,
     dataview=asn1_modules)]
 
+    # Generate the source file (.ads) header
+    ads_template = ['''\
+-- This file was generated automatically: DO NOT MODIFY IT !
+
+{dataview}
+
+package {process_name} is'''.format(process_name=process_name,
+    dataview=asn1_modules)]
+
+    # Generate the the code of the procedures
+    inner_procedures_code = []
+    for proc in process.inner_procedures:
+        proc_code, proc_local = inner_procedure(proc)
+        process_level_decl.extend(proc_local)
+        inner_procedures_code.extend(proc_code)
+
     # Generate the code for the process-level variable declarations
     taste_template.extend(process_level_decl)
+    
+    # Add the code of the procedures definitions
+    taste_template.extend(inner_procedures_code)
 
     # Generate the code for each input signal (provided interface)
-    for sig_name, sig_def in process.inputSignals.viewitems():
-        if sig_name == 'START':
+    for signal in process.input_signals:
+        if signal['name'] == 'START':
             continue
-        pi_header = 'procedure {sig_name}'.format(sig_name=sig_name)
-        if sig_def['in']:
-            param_name = sig_def['in'].keys()[0]
-            # Add PI parameter (only one is possible in TASTE PI)
+        pi_header = 'procedure {sig_name}'.format(sig_name=signal['name'])
+        param_name = signal.get('param_name') or 'MISSING_PARAM_NAME'
+        # Add (optional) PI parameter (only one is possible in TASTE PI)
+        if 'type' in signal:
             pi_header += '({pName}: access asn1Scc{pType})'.format(
-                    pName=param_name, pType=sig_def['in'][param_name]['type'])
+                pName=param_name, pType=signal['type'])
+
+        # Add declaration of the provided interface in the .ads file
+        ads_template.append('-- Provided interface "'+ signal['name'] + '"')
+        ads_template.append(pi_header + ';')
+
         pi_header += ' is'
         taste_template.append(pi_header)
         taste_template.append('begin')
@@ -775,17 +893,18 @@ package body {process_name} is'''.format(process_name=process_name,
             if state == 'START':
                 continue
             taste_template.append('when {state} =>'.format(state=state))
-            input_def = mapping[sig_name].get(state)
+            input_def = mapping[signal['name']].get(state)
             if input_def:
                 for inp in input_def.parameters:
                     # Assign the (optional and unique) parameter
                     # to the corresponding process variable
                     taste_template.append('l_{inp} := {tInp}.all;'.format(
-                        inp=inp, tInp=sig_def['in'].keys()[0]))
+                        inp=inp, tInp=param_name))
                 # Execute the correponding transition
                 if input_def.transition:
-                    taste_template.append('runTransition({t});'.format(
-                        t=input_def.inputlist[sig_name]))
+                    taste_template.append('runTransition({idx});'.format(
+                        idx=input_def.transition_id))
+                        #t=input_def.inputlist[signal['name']]))
                 else:
                     taste_template.append('null;')
             else:
@@ -793,9 +912,37 @@ package body {process_name} is'''.format(process_name=process_name,
         taste_template.append('when others =>')
         taste_template.append('null;')
         taste_template.append('end case;')
-        taste_template.append('end {sig_name};'.format(sig_name=sig_name))
+        taste_template.append('end {sig_name};'.format(sig_name=signal['name']))
         taste_template.append('\n')
 
+    # for the .ads file, generate the declaration of the required interfaces
+    # output signals are the asynchronous RI - only one parameter
+    for signal in process.output_signals:
+        ri_header = 'procedure {sig_name}'.format(sig_name=signal['name'])
+        param_name = signal.get('param_name') or 'MISSING_PARAM_NAME'
+        # Add (optional) RI parameter
+        if 'type' in signal:
+            ri_header += '({pName}: access asn1Scc{pType});'.format(
+                pName=param_name, pType=signal['type'])
+        ads_template.append('-- Required interface "'+ signal['name'] + '"')
+        ads_template.append(ri_header)
+        ads_template.append('pragma import(C, {sig}, "{proc}_RI_{sig}");'
+                .format(sig=signal['name'], proc=process_name))
+
+    # for the .ads file, generate the declaration of the external procedures
+    for proc in process.procedures:
+        ri_header = 'procedure {sig_name}'.format(sig_name=proc.inputString)
+        params = []
+        for param in proc.fpar:
+            params.append('{par[name]}: access asn1Scc{par[type]}'.format(
+                par=param))
+        if params:
+            ri_header += '(' + ';'.join(params) + ');'
+        ads_template.append('-- Sync required interface "'+ proc.inputString)
+        ads_template.append(ri_header)
+        ads_template.append('pragma import(C, {sig}, "{proc}_RI_{sig}");'
+                .format(sig=proc.inputString, proc=process_name))
+
     taste_template.append('procedure runTransition(trId: Integer) is')
 
     # Generate the code for all transitions
@@ -806,6 +953,13 @@ package body {process_name} is'''.format(process_name=process_name,
         code_transitions.append(code_tr)
         local_decl_transitions.extend(tr_local_decl)
 
+    # Generate code for the floating labels
+    code_labels = []
+    for label in process.floating_labels:
+        code_label, label_decl = floating_label(label)
+        code_labels.extend(code_label)
+        local_decl_transitions.extend(label_decl)
+
     # Declare the local variables needed by the transitions in the template
     decl = ['{line}'.format(line=l)
             for l in local_decl_transitions]
@@ -827,12 +981,22 @@ package body {process_name} is'''.format(process_name=process_name,
     taste_template.append('null;')
 
     taste_template.append('end case;')
+
+    # Add the code for the floating labels
+    taste_template.extend(code_labels)
+
     taste_template.append('end runTransition;')
     taste_template.append('\n')
 
     taste_template.extend(start_transition)
     taste_template.append('end {process_name};'
             .format(process_name=process_name))
+    
+    ads_template.append('end {process_name};'
+            .format(process_name=process_name))
 
-    with open(filename, 'w') as ada_file:
+    with open(process_name + '.adb', 'w') as ada_file:
         ada_file.write('\n'.join(format_ada_code(taste_template)))
+
+    with open(process_name + '.ads', 'w') as ada_file:
+        ada_file.write('\n'.join(format_ada_code(ads_template)))
